@@ -84,7 +84,6 @@ static size_t kpkt_len;
 /* Attributes for creating detached and joinable threads */
 pthread_attr_t th_attr;
 pthread_attr_t th_attr_detached;
-size_t detached_thread_stack_size = PTHREAD_STACK_MIN * 144;
 
 struct master_readmap_cond mrc = {
 	PTHREAD_MUTEX_INITIALIZER, PTHREAD_COND_INITIALIZER, 0, NULL, 0, 0, 0, 0};
@@ -242,15 +241,19 @@ int mkdir_path(const char *path, mode_t mode)
 int rmdir_path(struct autofs_point *ap, const char *path, dev_t dev)
 {
 	int len = strlen(path);
-	char buf[PATH_MAX];
+	char buf[PATH_MAX + 1];
 	char *cp;
 	int first = 1;
 	struct stat st;
 	struct statfs fs;
 
+	if (len > PATH_MAX) {
+		error(ap->logopt, "path longer than maximum length");
+		return -1;
+	}
 	strcpy(buf, path);
-	cp = buf + len;
 
+	cp = buf + len;
 	do {
 		*cp = '\0';
 
@@ -527,8 +530,11 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 	struct mapent_cache *mc;
 	struct mapent *me;
 	unsigned int is_mm_root = 0;
+	int cur_state;
 	int left;
 
+	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
+
 	me = lookup_source_mapent(ap, path, LKP_DISTINCT);
 	if (!me) {
 		char *ind_key;
@@ -542,50 +548,56 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 
 	if (me) {
 		mc = me->mc;
-		is_mm_root = (me->multi == me);
+		is_mm_root = IS_MM_ROOT(me);
 	}
 
 	left = 0;
 
-	if (me && me->multi) {
-		char root[PATH_MAX];
-		char *base;
-		int cur_state;
+	if (me && IS_MM(me)) {
+		char key[PATH_MAX + 1];
+		struct mapent *tmp;
+		int ret;
 
-		pthread_cleanup_push(cache_lock_cleanup, mc);
-
-		if (!strchr(me->multi->key, '/'))
-			/* Indirect multi-mount root */
-			/* sprintf okay - if it's mounted, it's
-			 * PATH_MAX or less bytes */
-			sprintf(root, "%s/%s", ap->path, me->multi->key);
-		else
-			strcpy(root, me->multi->key);
-
-		if (is_mm_root)
-			base = NULL;
-		else
-			base = me->key + strlen(root);
-
-		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-		/* Lock the closest parent nesting point for umount */
-		cache_multi_writelock(me->parent);
-		if (umount_multi_triggers(ap, me, root, base)) {
+		ret = tree_mapent_umount_offsets(me, 1);
+		if (!ret) {
 			warn(ap->logopt,
 			     "some offset mounts still present under %s", path);
 			left++;
 		}
-		cache_multi_unlock(me->parent);
+
+		if (me->len > PATH_MAX) {
+			crit(ap->logopt, "me->key too long for buffer");
+			return 1;
+		}
+
+		strcpy(key, me->key);
+
+		cache_unlock(mc);
+		cache_writelock(mc);
+		tmp = cache_lookup_distinct(mc, key);
+		/* mapent went away while we waited? */
+		if (tmp != me) {
+			cache_unlock(mc);
+			pthread_setcancelstate(cur_state, NULL);
+			return 0;
+		}
+
+		if (!left && IS_MM_ROOT(me)) {
+			if (!tree_mapent_delete_offsets(mc, me->key)) {
+				warn(ap->logopt, "couldn't delete offset list");
+				left++;
+			}
+		}
+
 		if (ap->entry->maps &&
 		    (ap->entry->maps->flags & MAP_FLAG_FORMAT_AMD))
 			cache_pop_mapent(me);
-		pthread_setcancelstate(cur_state, NULL);
-		pthread_cleanup_pop(0);
 	}
-
 	if (me)
 		cache_unlock(mc);
 
+	pthread_setcancelstate(cur_state, NULL);
+
 	if (left || is_autofs_fs)
 		return left;
 
@@ -653,7 +665,7 @@ int umount_multi(struct autofs_point *ap, const char *path, int incl)
 		/* Check if the autofs mount has browse mode enabled.
 		 * If so re-create the directory entry.
 		 */
-		if (ap->flags | MOUNT_FLAG_GHOST) {
+		if (ap->flags & MOUNT_FLAG_GHOST) {
 			int ret;
 
 			/* If the browse directory create fails log an
@@ -1108,7 +1120,7 @@ static int get_pkt(struct autofs_point *ap, union autofs_v5_packet_union *pkt)
 				estr = strerror_r(errno, buf, MAX_ERR_BUF);
 				error(ap->logopt,
 				      "read error on state pipe, "
-				      "read %u, error %s",
+				      "read %lu, error %s",
 				      read, estr);
 				st_mutex_unlock();
 				continue;
@@ -1126,7 +1138,7 @@ static int get_pkt(struct autofs_point *ap, union autofs_v5_packet_union *pkt)
 				estr = strerror_r(errno, buf, MAX_ERR_BUF);
 				error(ap->logopt,
 				      "read error on request pipe, "
-				      "read %u, expected %u error %s",
+				      "read %lu, expected %lu error %s",
 				       read, kpkt_len, estr);
 			}
 			return read;
@@ -1704,7 +1716,6 @@ void handle_mounts_startup_cond_destroy(void *arg)
 static void handle_mounts_cleanup(void *arg)
 {
 	struct autofs_point *ap;
-	char path[PATH_MAX + 1];
 	char buf[MAX_ERR_BUF];
 	unsigned int clean = 0, submount, logopt;
 	unsigned int pending = 0;
@@ -1714,7 +1725,6 @@ static void handle_mounts_cleanup(void *arg)
 	logopt = ap->logopt;
 	submount = ap->submount;
 
-	strcpy(path, ap->path);
 	if (!submount && strcmp(ap->path, "/-") &&
 	    ap->flags & MOUNT_FLAG_DIR_CREATED)
 		clean = 1;
@@ -1739,8 +1749,8 @@ static void handle_mounts_cleanup(void *arg)
 	/* Don't signal the handler if we have already done so */
 	if (!list_empty(&master_list->completed))
 		pending = 1;
-	master_remove_mapent(ap->entry);
-	master_source_unlock(ap->entry);
+
+	info(logopt, "shut down path %s", ap->path);
 
 	destroy_logpri_fifo(ap);
 
@@ -1750,21 +1760,21 @@ static void handle_mounts_cleanup(void *arg)
 	 * here.
 	 */
 	if (submount) {
-		mounts_mutex_unlock(ap->parent);
 		master_source_unlock(ap->parent->entry);
 		master_free_mapent_sources(ap->entry, 1);
 		master_free_mapent(ap->entry);
 	}
 
 	if (clean) {
-		if (rmdir(path) == -1) {
+		if (rmdir(ap->path) == -1) {
 			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 			warn(logopt, "failed to remove dir %s: %s",
-			     path, estr);
+			     ap->path, estr);
 		}
 	}
 
-	info(logopt, "shut down path %s", path);
+	master_remove_mapent(ap->entry);
+	master_source_unlock(ap->entry);
 
 	/*
 	 * If we are not a submount send a signal to the signal handler
@@ -1788,13 +1798,9 @@ static int submount_source_writelock_nested(struct autofs_point *ap)
 	if (status)
 		goto done;
 
-	mounts_mutex_lock(parent);
-
 	status = pthread_rwlock_trywrlock(&ap->entry->source_lock);
-	if (status) {
-		mounts_mutex_unlock(parent);
+	if (status)
 		master_source_unlock(parent->entry);
-	}
 
 done:
 	if (status && status != EBUSY) {
@@ -1810,7 +1816,6 @@ static void submount_source_unlock_nested(struct autofs_point *ap)
 	struct autofs_point *parent = ap->parent;
 
 	master_source_unlock(ap->entry);
-	mounts_mutex_unlock(parent);
 	master_source_unlock(parent->entry);
 }
 
@@ -1924,6 +1929,8 @@ void *handle_mounts(void *arg)
 	status = pthread_mutex_lock(&suc->mutex);
 	if (status) {
 		logerr("failed to lock startup condition mutex!");
+		if (root)
+			free(root);
 		fatal(status);
 	}
 
@@ -2274,7 +2281,7 @@ int main(int argc, char *argv[])
 	time_t timeout;
 	time_t age = monotonic_time(NULL);
 	struct rlimit rlim;
-	const char *options = "+hp:t:vmdD:SfVrO:l:n:CFUM";
+	const char *options = "+hp:t:vmdD:SfVrO:l:n:CFUM:";
 	static const struct option long_options[] = {
 		{"help", 0, 0, 'h'},
 		{"pid-file", 1, 0, 'p'},
@@ -2615,34 +2622,6 @@ int main(int argc, char *argv[])
 		exit(1);
 	}
 
-#ifdef _POSIX_THREAD_ATTR_STACKSIZE
-	if (pthread_attr_setstacksize(
-			&th_attr_detached, detached_thread_stack_size)) {
-		logerr("%s: failed to set stack size thread attribute!",
-		       program);
-		if (start_pipefd[1] != -1) {
-			res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
-			close(start_pipefd[1]);
-		}
-		release_flag_file();
-		macro_free_global_table();
-		exit(1);
-	}
-#endif
-
-	if (pthread_attr_getstacksize(
-			&th_attr_detached, &detached_thread_stack_size)) {
-		logerr("%s: failed to get detached thread stack size!",
-		       program);
-		if (start_pipefd[1] != -1) {
-			res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
-			close(start_pipefd[1]);
-		}
-		release_flag_file();
-		macro_free_global_table();
-		exit(1);
-	}
-
 	info(logging, "Starting automounter version %s, master map %s",
 		version, master_list->name);
 	info(logging, "using kernel protocol version %d.%02d",
