@@ -84,11 +84,27 @@ static void mnts_cleanup(void *arg)
 int do_umount_autofs_direct(struct autofs_point *ap, struct mapent *me)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
+	struct mapent_cache *mc = me->mc;
 	char buf[MAX_ERR_BUF];
 	int ioctlfd = -1, rv, left, retries;
+	char key[PATH_MAX + 1];
+	struct mapent *tmp;
 	int opened = 0;
 
-	left = umount_multi(ap, me->key, 0);
+	if (me->len > PATH_MAX) {
+		error(ap->logopt, "path too long");
+		return 1;
+	}
+	strcpy(key, me->key);
+
+	cache_unlock(mc);
+	left = umount_multi(ap, key, 0);
+	cache_readlock(mc);
+	tmp = cache_lookup_distinct(mc, key);
+	if (tmp != me) {
+		error(ap->logopt, "key %s no longer in mapent cache", key);
+		return -1;
+	}
 	if (left) {
 		warn(ap->logopt, "could not unmount %d dirs under %s",
 		     left, me->key);
@@ -150,7 +166,7 @@ int do_umount_autofs_direct(struct autofs_point *ap, struct mapent *me)
 
 	retries = UMOUNT_RETRIES;
 	while ((rv = umount(me->key)) == -1 && retries--) {
-		struct timespec tm = {0, 200000000};
+		struct timespec tm = {0, 50000000};
 		if (errno != EBUSY)
 			break;
 		nanosleep(&tm, NULL);
@@ -213,6 +229,7 @@ int umount_autofs_direct(struct autofs_point *ap)
 		mc = map->mc;
 		pthread_cleanup_push(cache_lock_cleanup, mc);
 		cache_readlock(mc);
+restart:
 		me = cache_enumerate(mc, NULL);
 		while (me) {
 			int error;
@@ -230,6 +247,9 @@ int umount_autofs_direct(struct autofs_point *ap)
 			 * failed umount.
 			 */
 			error = do_umount_autofs_direct(ap, me);
+			/* cache became invalid, restart */
+			if (error == -1)
+				goto restart;
 			if (!error)
 				goto done;
 
@@ -573,7 +593,7 @@ int umount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 
 	retries = UMOUNT_RETRIES;
 	while ((rv = umount(me->key)) == -1 && retries--) {
-		struct timespec tm = {0, 200000000};
+		struct timespec tm = {0, 50000000};
 		if (errno != EBUSY)
 			break;
 		nanosleep(&tm, NULL);
@@ -605,13 +625,10 @@ force_umount:
 	} else
 		info(ap->logopt, "umounted offset mount %s", me->key);
 
-	if (!rv)
-		mnts_remove_mount(me->key, MNTS_OFFSET);
-
 	return rv;
 }
 
-int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *root, const char *offset)
+int mount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 {
 	const char *str_offset = mount_type_str(t_offset);
 	struct ioctl_ops *ops = get_ioctl_ops();
@@ -623,7 +640,6 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 	const char *hosts_map_name = "-hosts";
 	const char *map_name = hosts_map_name;
 	const char *type;
-	char mountpoint[PATH_MAX];
 	struct mnt_list *mnt;
 
 	if (ops->version && ap->flags & MOUNT_FLAG_REMOUNT) {
@@ -681,11 +697,8 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 			return MOUNT_OFFSET_OK;
 	}
 
-	strcpy(mountpoint, root);
-	strcat(mountpoint, offset);
-
 	/* In case the directory doesn't exist, try to mkdir it */
-	if (mkdir_path(mountpoint, mp_mode) < 0) {
+	if (mkdir_path(me->key, mp_mode) < 0) {
 		if (errno == EEXIST) {
 			/*
 			 * If the mount point directory is a real mount
@@ -693,8 +706,8 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 			 * a mount that has been automatically mounted by
 			 * the kernel NFS client.
 			 */
-			if (me->multi != me &&
-			    is_mounted(mountpoint, MNTS_REAL))
+			if (!IS_MM_ROOT(me) &&
+			    is_mounted(me->key, MNTS_REAL))
 				return MOUNT_OFFSET_IGNORE;
 
 			/* 
@@ -714,13 +727,13 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 			debug(ap->logopt,
 			     "can't create mount directory: %s, %s",
-			     mountpoint, estr);
+			     me->key, estr);
 			return MOUNT_OFFSET_FAIL;
 		} else {
 			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 			crit(ap->logopt,
 			     "failed to create mount directory: %s, %s",
-			     mountpoint, estr);
+			     me->key, estr);
 			return MOUNT_OFFSET_FAIL;
 		}
 	} else {
@@ -730,56 +743,51 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 
 	debug(ap->logopt,
 	      "calling mount -t autofs " SLOPPY " -o %s automount %s",
-	      mp->options, mountpoint);
+	      mp->options, me->key);
 
 	type = ap->entry->maps->type;
 	if (!type || strcmp(ap->entry->maps->type, "hosts"))
 		map_name = me->mc->map->argv[0];
 
-	ret = mount(map_name, mountpoint, "autofs", MS_MGC_VAL, mp->options);
+	ret = mount(map_name, me->key, "autofs", MS_MGC_VAL, mp->options);
 	if (ret) {
 		crit(ap->logopt,
 		     "failed to mount offset trigger %s at %s",
-		     me->key, mountpoint);
+		     me->key, me->key);
 		goto out_err;
 	}
 
-	ret = stat(mountpoint, &st);
+	ret = stat(me->key, &st);
 	if (ret == -1) {
+		int save_errno = errno;
+
 		error(ap->logopt,
-		     "failed to stat direct mount trigger %s", mountpoint);
-		goto out_umount;
+		     "failed to stat direct mount trigger %s", me->key);
+		if (save_errno != ENOENT)
+			goto out_umount;
+		goto out_err;
 	}
 
-	ops->open(ap->logopt, &ioctlfd, st.st_dev, mountpoint);
+	ops->open(ap->logopt, &ioctlfd, st.st_dev, me->key);
 	if (ioctlfd < 0) {
-		crit(ap->logopt, "failed to create ioctl fd for %s", mountpoint);
+		crit(ap->logopt, "failed to create ioctl fd for %s", me->key);
 		goto out_umount;
 	}
 
 	ops->timeout(ap->logopt, ioctlfd, timeout);
 	cache_set_ino_index(me->mc, me->key, st.st_dev, st.st_ino);
-	if (ap->logopt & LOGOPT_DEBUG)
-		notify_mount_result(ap, mountpoint, timeout, str_offset);
-	else
-		notify_mount_result(ap, me->key, timeout, str_offset);
+	notify_mount_result(ap, me->key, timeout, str_offset);
 	ops->close(ap->logopt, ioctlfd);
 
-	mnt = mnts_add_mount(ap, mountpoint, MNTS_OFFSET);
-	if (!mnt)
-		error(ap->logopt,
-		      "failed to add offset mount %s to mounted list",
-		      mountpoint);
-
-	debug(ap->logopt, "mounted trigger %s at %s", me->key, mountpoint);
+	debug(ap->logopt, "mounted trigger %s", me->key);
 
 	return MOUNT_OFFSET_OK;
 
 out_umount:
-	umount(mountpoint);
+	umount(me->key);
 out_err:
-	if (stat(mountpoint, &st) == 0 && me->flags & MOUNT_FLAG_DIR_CREATED)
-		 rmdir_path(ap, mountpoint, st.st_dev);
+	if (stat(me->key, &st) == 0 && me->flags & MOUNT_FLAG_DIR_CREATED)
+		 rmdir_path(ap, me->key, st.st_dev);
 
 	return MOUNT_OFFSET_FAIL;
 }
@@ -888,7 +896,7 @@ cont:
 			ioctlfd = me->ioctlfd;
 
 			ret = ops->expire(ap->logopt, ioctlfd, mnt->mp, how);
-			if (ret) {
+			if (ret == 1) {
 				left++;
 				pthread_setcancelstate(cur_state, NULL);
 				continue;
@@ -914,7 +922,7 @@ cont:
 
 		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
 		ret = ops->expire(ap->logopt, ioctlfd, mnt->mp, how);
-		if (ret)
+		if (ret == 1)
 			left++;
 		pthread_setcancelstate(cur_state, NULL);
 	}
@@ -1031,13 +1039,18 @@ int handle_packet_expire_direct(struct autofs_point *ap, autofs_packet_expire_di
 		map = map->next;
 	}
 
-	if (!me) {
+	if (!me || me->len >= PATH_MAX) {
 		/*
 		 * Shouldn't happen as we have been sent this following
 		 * successful thread creation and lookup.
 		 */
-		crit(ap->logopt, "can't find map entry for (%lu,%lu)",
-		    (unsigned long) pkt->dev, (unsigned long) pkt->ino);
+		if (!me)
+			crit(ap->logopt, "can't find map entry for (%lu,%lu)",
+			    (unsigned long) pkt->dev, (unsigned long) pkt->ino);
+		else {
+			cache_unlock(mc);
+			crit(ap->logopt, "lookup key is too long");
+		}
 		master_source_unlock(ap->entry);
 		pthread_setcancelstate(state, NULL);
 		return 1;
@@ -1083,7 +1096,6 @@ int handle_packet_expire_direct(struct autofs_point *ap, autofs_packet_expire_di
 	mt->ap = ap;
 	mt->ioctlfd = me->ioctlfd;
 	mt->mc = mc;
-	/* TODO: check length here */
 	strcpy(mt->name, me->key);
 	mt->dev = me->dev;
 	mt->type = NFY_EXPIRE;
@@ -1218,6 +1230,7 @@ static void *do_mount_direct(void *arg)
 		struct mapent *me;
 		struct statfs fs;
 		unsigned int close_fd = 0;
+		unsigned int flags = MNTS_DIRECT|MNTS_MOUNTED;
 
 		sbmnt = mnts_find_submount(mt.name);
 		if (statfs(mt.name, &fs) == -1 ||
@@ -1232,10 +1245,12 @@ static void *do_mount_direct(void *arg)
 			 * for direct mount multi-mounts with no real mount at
 			 * their base so they will be expired.
 			 */
-			if (close_fd && me == me->multi)
+			if (close_fd && IS_MM_ROOT(me))
 				close_fd = 0;
 			if (!close_fd)
 				me->ioctlfd = mt.ioctlfd;
+			if (IS_MM(me) && !IS_MM_ROOT(me))
+				flags |= MNTS_OFFSET;
 		}
 		ops->send_ready(ap->logopt, mt.ioctlfd, mt.wait_queue_token);
 		cache_unlock(mt.mc);
@@ -1244,7 +1259,7 @@ static void *do_mount_direct(void *arg)
 
 		info(ap->logopt, "mounted %s", mt.name);
 
-		mnts_set_mounted_mount(ap, mt.name);
+		mnts_set_mounted_mount(ap, mt.name, flags);
 
 		conditional_alarm_add(ap, ap->exp_runfreq);
 	} else {
