@@ -748,7 +748,7 @@ int lookup_ghost(struct autofs_point *ap, const char *root)
 					goto next;
 
 				/* It's a busy multi-mount - leave till next time */
-				if (list_empty(&me->multi_list))
+				if (IS_MM(me))
 					error(ap->logopt,
 					      "invalid key %s", me->key);
 				goto next;
@@ -838,12 +838,12 @@ static int lookup_amd_instance(struct autofs_point *ap,
 	char *m_key;
 
 	me = cache_lookup_distinct(map->mc, name);
-	if (!me || !me->multi) {
+	if (!me || !IS_MM(me)) {
 		error(ap->logopt, "expected multi mount entry not found");
 		return NSS_STATUS_UNKNOWN;
 	}
 
-	m_key = malloc(strlen(ap->path) + strlen(me->multi->key) + 2);
+	m_key = malloc(ap->len + MM_ROOT(me)->len + 2);
 	if (!m_key) {
 		error(ap->logopt,
 		     "failed to allocate storage for search key");
@@ -852,7 +852,7 @@ static int lookup_amd_instance(struct autofs_point *ap,
 
 	strcpy(m_key, ap->path);
 	strcat(m_key, "/");
-	strcat(m_key, me->multi->key);
+	strcat(m_key, MM_ROOT(me)->key);
 
 	mnt = mnts_find_amdmount(m_key);
 	free(m_key);
@@ -1355,7 +1355,7 @@ void lookup_prune_one_cache(struct autofs_point *ap, struct mapent_cache *mc, ti
 			 * created on demand and managed by expire and don't
 			 * prune the multi-map owner map entry.
 			 */
-			if (*me->key == '/' || me->multi == me) {
+			if (*me->key == '/' || IS_MM_ROOT(me)) {
 				me = cache_enumerate(mc, me);
 				continue;
 			}
@@ -1375,11 +1375,11 @@ void lookup_prune_one_cache(struct autofs_point *ap, struct mapent_cache *mc, ti
 		}
 
 		key = strdup(me->key);
-		me = cache_enumerate(mc, me);
 		/* Don't consider any entries with a wildcard */
 		if (!key || strchr(key, '*')) {
 			if (key)
 				free(key);
+			me = cache_enumerate(mc, me);
 			continue;
 		}
 
@@ -1387,6 +1387,7 @@ void lookup_prune_one_cache(struct autofs_point *ap, struct mapent_cache *mc, ti
 		if (!path) {
 			warn(ap->logopt, "can't malloc storage for path");
 			free(key);
+			me = cache_enumerate(mc, me);
 			continue;
 		}
 
@@ -1414,14 +1415,16 @@ void lookup_prune_one_cache(struct autofs_point *ap, struct mapent_cache *mc, ti
 		}
 		if (!valid &&
 		    is_mounted(path, MNTS_REAL)) {
-			debug(ap->logopt, "prune posponed, %s mounted", path);
+			debug(ap->logopt, "prune postponed, %s mounted", path);
 			free(key);
 			free(path);
+			me = cache_enumerate(mc, me);
 			continue;
 		}
 		if (valid)
 			cache_unlock(valid->mc);
 
+		me = cache_enumerate(mc, me);
 		if (me)
 			next_key = strdup(me->key);
 
@@ -1456,6 +1459,24 @@ void lookup_prune_one_cache(struct autofs_point *ap, struct mapent_cache *mc, ti
 next:
 		cache_readlock(mc);
 		if (next_key) {
+			/* The lock release and reaquire above can mean
+			 * a number of things could happen.
+			 *
+			 * First, mapents could be added between the
+			 * current mapent and the mapent of next_key.
+			 * Don't care about that because there's no
+			 * need to prune newly added entries.
+			 *
+			 * Second, the next mapent data could have
+			 * changed. Don't care about that either since
+			 * we are looking to prune stale map entries
+			 * and don't care when they become stale.
+			 *
+			 * Finally, the mapent of next_key could have
+			 * gone away. Again don't care about this either,
+			 * the loop will exit prematurely so just wait
+			 * until the next prune and try again.
+			 */
 			me = cache_lookup_distinct(mc, next_key);
 			free(next_key);
 		}
