@@ -31,10 +31,9 @@
 #include "master.h"
 #include "macros.h"
 #include "log.h"
+#include "mounts.h"
 #include "rpc_subs.h"
-#include "mounts.h"
 #include "parse_subs.h"
-#include "mounts.h"
 #include "dev-ioctl-lib.h"
 #include "parse_amd.h"
 
@@ -141,7 +140,7 @@ struct autofs_point;
 #define NULL_MAP_HASHSIZE	64
 #define NEGATIVE_TIMEOUT	10
 #define POSITIVE_TIMEOUT	120
-#define UMOUNT_RETRIES		8
+#define UMOUNT_RETRIES		16
 #define EXPIRE_RETRIES		3
 
 struct mapent_cache {
@@ -163,15 +162,16 @@ struct stack {
 struct mapent {
 	struct mapent *next;
 	struct list_head ino_index;
-	pthread_rwlock_t multi_rwlock;
-	struct list_head multi_list;
 	struct mapent_cache *mc;
 	struct map_source *source;
 	/* Need to know owner if we're a multi-mount */
-	struct mapent *multi;
+	struct tree_node *mm_root;
 	/* Parent nesting point within multi-mount */
-	struct mapent *parent;
+	struct tree_node *mm_parent;
+	struct tree_node node;
+	struct list_head work;
 	char *key;
+	size_t len;
 	char *mapent;
 	struct stack *stack;
 	time_t age;
@@ -185,6 +185,11 @@ struct mapent {
 	ino_t ino;
 };
 
+#define IS_MM(me)	(me->mm_root)
+#define IS_MM_ROOT(me)	(me->mm_root == &me->node)
+#define MM_ROOT(me)	(MAPENT(me->mm_root))
+#define MM_PARENT(me)	(MAPENT(me->mm_parent))
+
 void cache_lock_cleanup(void *arg);
 void cache_readlock(struct mapent_cache *mc);
 void cache_writelock(struct mapent_cache *mc);
@@ -202,26 +207,19 @@ struct mapent *cache_lookup_next(struct mapent_cache *mc, struct mapent *me);
 struct mapent *cache_lookup_key_next(struct mapent *me);
 struct mapent *cache_lookup(struct mapent_cache *mc, const char *key);
 struct mapent *cache_lookup_distinct(struct mapent_cache *mc, const char *key);
-struct mapent *cache_lookup_offset(const char *prefix, const char *offset, int start, struct list_head *head);
 struct mapent *cache_partial_match(struct mapent_cache *mc, const char *prefix);
 struct mapent *cache_partial_match_wild(struct mapent_cache *mc, const char *prefix);
 int cache_add(struct mapent_cache *mc, struct map_source *ms, const char *key, const char *mapent, time_t age);
 int cache_update_offset(struct mapent_cache *mc, const char *mkey, const char *key, const char *mapent, time_t age);
 int cache_lookup_negative(struct mapent *me, const char *key);
 void cache_update_negative(struct mapent_cache *mc, struct map_source *ms, const char *key, time_t timeout);
-int cache_set_parents(struct mapent *mm);
+struct mapent *cache_get_offset_parent(struct mapent_cache *mc, const char *key);
 int cache_update(struct mapent_cache *mc, struct map_source *ms, const char *key, const char *mapent, time_t age);
 int cache_delete(struct mapent_cache *mc, const char *key);
-int cache_delete_offset(struct mapent_cache *mc, const char *key);
-void cache_multi_readlock(struct mapent *me);
-void cache_multi_writelock(struct mapent *me);
-void cache_multi_unlock(struct mapent *me);
-int cache_delete_offset_list(struct mapent_cache *mc, const char *key);
 void cache_release(struct map_source *map);
 void cache_clean_null_cache(struct mapent_cache *mc);
 void cache_release_null_cache(struct master *master);
 struct mapent *cache_enumerate(struct mapent_cache *mc, struct mapent *me);
-char *cache_get_offset(const char *prefix, char *offset, int start, struct list_head *head, struct list_head **pos);
 
 /* Utility functions */
 
@@ -548,6 +546,7 @@ struct kernel_mod_version {
 struct autofs_point {
 	pthread_t thid;
 	char *path;			/* Mount point name */
+	size_t len;			/* Length of mount point name */
 	mode_t mode;			/* Mount point mode */
 	char *pref;			/* amd prefix */
 	int pipefd;			/* File descriptor for pipe */
@@ -567,7 +566,6 @@ struct autofs_point {
 	enum states state;		/* Current state */
 	int state_pipe[2];		/* State change router pipe */
 	struct autofs_point *parent;	/* Owner of mounts list for submount */
-	pthread_mutex_t mounts_mutex;	/* Protect mount lists */
 	struct list_head mounts;	/* List of autofs mounts at current level */
 	unsigned int submount;		/* Is this a submount */
 	unsigned int submnt_count;	/* Number of submounts */
@@ -597,7 +595,7 @@ int expire_offsets_direct(struct autofs_point *ap, struct mapent *me, int now);
 int mount_autofs_indirect(struct autofs_point *ap, const char *root);
 int do_mount_autofs_direct(struct autofs_point *ap, struct mapent *me, time_t timeout);
 int mount_autofs_direct(struct autofs_point *ap);
-int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *root, const char *offset);
+int mount_autofs_offset(struct autofs_point *ap, struct mapent *me);
 void submount_signal_parent(struct autofs_point *ap, unsigned int success);
 void close_mount_fds(struct autofs_point *ap);
 int umount_autofs_indirect(struct autofs_point *ap, const char *root);
