@@ -63,6 +63,7 @@ struct master {
 	unsigned int depth;
 	unsigned int reading;
 	unsigned int read_fail;
+	unsigned int readall;
 	unsigned int default_ghost;
 	unsigned int default_logging;
 	unsigned int default_timeout;
@@ -108,24 +109,20 @@ void master_source_current_wait(struct master_mapent *);
 void master_source_current_signal(struct master_mapent *);
 struct master_mapent *master_find_mapent(struct master *, const char *);
 unsigned int master_partial_match_mapent(struct master *, const char *);
-struct autofs_point *__master_find_submount(struct autofs_point *, const char *);
-struct autofs_point *master_find_submount(struct autofs_point *, const char *);
-struct amd_entry *__master_find_amdmount(struct autofs_point *, const char *);
-struct amd_entry *master_find_amdmount(struct autofs_point *, const char *);
 struct master_mapent *master_new_mapent(struct master *, const char *, time_t);
 void master_add_mapent(struct master *, struct master_mapent *);
 void master_remove_mapent(struct master_mapent *);
 void master_free_mapent_sources(struct master_mapent *, unsigned int);
 void master_free_mapent(struct master_mapent *);
 struct master *master_new(const char *, unsigned int, unsigned int);
-int master_read_master(struct master *, time_t, int);
-int master_submount_list_empty(struct autofs_point *ap);
+int master_read_master(struct master *, time_t);
 int master_notify_submount(struct autofs_point *, const char *path, enum states);
 void master_notify_state_change(struct master *, int);
-int master_mount_mounts(struct master *, time_t, int);
+int master_mount_mounts(struct master *, time_t);
 int dump_map(struct master *, const char *, const char *);
 int master_show_mounts(struct master *);
 unsigned int master_get_logopt(void);
+int __master_list_empty(struct master *);
 int master_list_empty(struct master *);
 int master_done(struct master *);
 int master_kill(struct master *);
