@@ -24,34 +24,6 @@
 
 #include "automount.h"
 
-void cache_dump_multi(struct list_head *list)
-{
-	struct list_head *p;
-	struct mapent *me;
-
-	list_for_each(p, list) {
-		me = list_entry(p, struct mapent, multi_list);
-		logmsg("key=%s", me->key);
-	}
-}
-
-void cache_dump_cache(struct mapent_cache *mc)
-{
-	struct mapent *me;
-	unsigned int i;
-
-	for (i = 0; i < mc->size; i++) {
-		me = mc->hash[i];
-		if (me == NULL)
-			continue;
-		while (me) {
-			logmsg("me->key=%s me->multi=%p dev=%ld ino=%ld",
-				me->key, me->multi, me->dev, me->ino);
-			me = me->next;
-		}
-	}
-}
-
 void cache_readlock(struct mapent_cache *mc)
 {
 	int status;
@@ -108,58 +80,6 @@ void cache_lock_cleanup(void *arg)
 	return;
 }
 
-void cache_multi_readlock(struct mapent *me)
-{
-	int status;
-
-	if (!me)
-		return;
-
-	status = pthread_rwlock_rdlock(&me->multi_rwlock);
-	if (status) {
-		logmsg("mapent cache multi mutex lock failed");
-		fatal(status);
-	}
-	return;
-}
-
-void cache_multi_writelock(struct mapent *me)
-{
-	int status;
-
-	if (!me)
-		return;
-
-	status = pthread_rwlock_wrlock(&me->multi_rwlock);
-	if (status) {
-		logmsg("mapent cache multi mutex lock failed");
-		fatal(status);
-	}
-	return;
-}
-
-void cache_multi_unlock(struct mapent *me)
-{
-	int status;
-
-	if (!me)
-		return;
-
-	status = pthread_rwlock_unlock(&me->multi_rwlock);
-	if (status) {
-		logmsg("mapent cache multi mutex unlock failed");
-		fatal(status);
-	}
-	return;
-}
-
-void cache_multi_lock_cleanup(void *arg)
-{
-	struct mapent *me = (struct mapent *) arg;
-	cache_multi_unlock(me);
-	return;
-}
-
 static inline void ino_index_lock(struct mapent_cache *mc)
 {
 	int status = pthread_mutex_lock(&mc->ino_index_mutex);
@@ -426,7 +346,7 @@ struct mapent *cache_lookup_first(struct mapent_cache *mc)
 
 		while (me) {
 			/* Multi mount entries are not primary */
-			if (me->multi && me->multi != me) {
+			if (IS_MM(me) && !IS_MM_ROOT(me)) {
 				me = me->next;
 				continue;
 			}
@@ -449,7 +369,7 @@ struct mapent *cache_lookup_next(struct mapent_cache *mc, struct mapent *me)
 	this = me->next;
 	while (this) {
 		/* Multi mount entries are not primary */
-		if (this->multi && this->multi != this) {
+		if (IS_MM(this) && !IS_MM_ROOT(this)) {
 			this = this->next;
 			continue;
 		}
@@ -465,7 +385,7 @@ struct mapent *cache_lookup_next(struct mapent_cache *mc, struct mapent *me)
 
 			while (this) {
 				/* Multi mount entries are not primary */
-				if (this->multi && this->multi != this) {
+				if (IS_MM(this) && !IS_MM_ROOT(this)) {
 					this = this->next;
 					continue;
 				}
@@ -487,7 +407,7 @@ struct mapent *cache_lookup_key_next(struct mapent *me)
 	next = me->next;
 	while (next) {
 		/* Multi mount entries are not primary */
-		if (me->multi && me->multi != me)
+		if (IS_MM(me) && !IS_MM_ROOT(me))
 			continue;
 		if (!strcmp(me->key, next->key))
 			return next;
@@ -541,30 +461,6 @@ struct mapent *cache_lookup_distinct(struct mapent_cache *mc, const char *key)
 	return NULL;
 }
 
-/* Lookup an offset within a multi-mount entry */
-struct mapent *cache_lookup_offset(const char *prefix, const char *offset, int start, struct list_head *head)
-{
-	struct list_head *p;
-	struct mapent *this;
-	/* Keys for direct maps may be as long as a path name */
-	char o_key[PATH_MAX];
-	/* Avoid "//" at the beginning of paths */
-	const char *path_prefix = strlen(prefix) > 1 ? prefix : "";
-	size_t size;
-
-	/* root offset duplicates "/" */
-	size = snprintf(o_key, sizeof(o_key), "%s%s", path_prefix, offset);
-	if (size >= sizeof(o_key))
-		return NULL;
-
-	list_for_each(p, head) {
-		this = list_entry(p, struct mapent, multi_list);
-		if (!strcmp(&this->key[start], o_key))
-			return this;
-	}
-	return NULL;
-}
-
 /* cache must be read locked by caller */
 static struct mapent *__cache_partial_match(struct mapent_cache *mc,
 					    const char *prefix,
@@ -626,18 +522,21 @@ int cache_add(struct mapent_cache *mc, struct map_source *ms, const char *key, c
 	struct mapent *me, *existing = NULL;
 	char *pkey, *pent;
 	u_int32_t hashval = hash(key, mc->size);
-	int status;
+	size_t len;
 
 	me = (struct mapent *) malloc(sizeof(struct mapent));
 	if (!me)
 		return CHE_FAIL;
 
-	pkey = malloc(strlen(key) + 1);
+	len = strlen(key);
+
+	pkey = malloc(len + 1);
 	if (!pkey) {
 		free(me);
 		return CHE_FAIL;
 	}
 	me->key = strcpy(pkey, key);
+	me->len = len;
 
 	if (mapent) {
 		pent = malloc(strlen(mapent) + 1);
@@ -656,19 +555,16 @@ int cache_add(struct mapent_cache *mc, struct map_source *ms, const char *key, c
 	me->status = 0;
 	me->mc = mc;
 	me->source = ms;
+	me->mm_root = NULL;
+	me->mm_parent = NULL;
+	INIT_TREE_NODE(&me->node);
 	INIT_LIST_HEAD(&me->ino_index);
-	INIT_LIST_HEAD(&me->multi_list);
-	me->multi = NULL;
-	me->parent = NULL;
+	INIT_LIST_HEAD(&me->work);
 	me->ioctlfd = -1;
 	me->dev = (dev_t) -1;
 	me->ino = (ino_t) -1;
 	me->flags = 0;
 
-	status = pthread_rwlock_init(&me->multi_rwlock, NULL);
-	if (status)
-		fatal(status);
-
 	/* 
 	 * We need to add to the end if values exist in order to
 	 * preserve the order in which the map was read on lookup.
@@ -693,33 +589,6 @@ int cache_add(struct mapent_cache *mc, struct map_source *ms, const char *key, c
 	return CHE_OK;
 }
 
-/* cache must be write locked by caller */
-static void cache_add_ordered_offset(struct mapent *me, struct list_head *head)
-{
-	struct list_head *p;
-	struct mapent *this;
-
-	list_for_each(p, head) {
-		size_t tlen;
-		int eq;
-
-		this = list_entry(p, struct mapent, multi_list);
-		tlen = strlen(this->key);
-
-		eq = strncmp(this->key, me->key, tlen);
-		if (!eq && tlen == strlen(me->key))
-			return;
-
-		if (eq > 0) {
-			list_add_tail(&me->multi_list, p);
-			return;
-		}
-	}
-	list_add_tail(&me->multi_list, p);
-
-	return;
-}
-
 /* cache must be write locked by caller */
 int cache_update_offset(struct mapent_cache *mc, const char *mkey, const char *key, const char *mapent, time_t age)
 {
@@ -760,14 +629,6 @@ int cache_update_offset(struct mapent_cache *mc, const char *mkey, const char *k
 		return CHE_FAIL;
 	}
 
-	me = cache_lookup_distinct(mc, key);
-	if (me) {
-		cache_add_ordered_offset(me, &owner->multi_list);
-		me->multi = owner;
-		goto done;
-	}
-	ret = CHE_FAIL;
-done:
 	return ret; 
 }
 
@@ -827,54 +688,40 @@ void cache_update_negative(struct mapent_cache *mc,
 }
 
 
-static struct mapent *get_parent(const char *key, struct list_head *head, struct list_head **pos)
+struct mapent *cache_get_offset_parent(struct mapent_cache *mc, const char *key)
 {
-	struct list_head *next;
-	struct mapent *this, *last;
-	int eq;
+	struct mapent *me;
+	char *parent, *tail;
+	int key_len;
 
-	last = NULL;
-	next = *pos ? (*pos)->next : head->next;
+	key_len = strlen(key);
 
-	list_for_each(next, head) {
-		this = list_entry(next, struct mapent, multi_list);
+	/* Check if this is the root offset */
+	if (key[key_len - 1] == '/')
+		return NULL;
 
-		if (!strcmp(this->key, key))
+	parent = strdup(key);
+	tail = &parent[key_len - 1];
+
+	while (*tail) {
+		while (*tail != '/')
+			tail--;
+
+		*tail = 0;
+
+		tail--;
+		if (tail == parent)
 			break;
 
-		eq = strncmp(this->key, key, strlen(this->key));
-		if (eq == 0) {
-			*pos = next;
-			last = this;
-			continue;
+		me = cache_lookup_distinct(mc, parent);
+		if (me) {
+			free(parent);
+			return me;
 		}
 	}
+	free(parent);
 
-	return last;
-}
-
-int cache_set_parents(struct mapent *mm)
-{
-	struct list_head *multi_head, *p, *pos;
-	struct mapent *this;
-
-	if (!mm->multi)
-		return 0;
-
-	pos = NULL;
-	multi_head = &mm->multi->multi_list;
-
-	list_for_each(p, multi_head) {
-		struct mapent *parent;
-		this = list_entry(p, struct mapent, multi_list);
-		parent = get_parent(this->key, multi_head, &pos);
-		if (parent)
-			this->parent = parent;
-		else
-			this->parent = mm->multi;
-	}
-
-	return 1;
+	return NULL;
 }
 
 /* cache must be write locked by caller */
@@ -918,82 +765,27 @@ int cache_update(struct mapent_cache *mc, struct map_source *ms, const char *key
 	return ret;
 }
 
-/* cache_multi_lock of the multi mount owner must be held by caller */
-int cache_delete_offset(struct mapent_cache *mc, const char *key)
-{
-	u_int32_t hashval = hash(key, mc->size);
-	struct mapent *me = NULL, *pred;
-	int status;
-
-	me = mc->hash[hashval];
-	if (!me)
-		return CHE_FAIL;
-
-	if (strcmp(key, me->key) == 0) {
-		if (me->multi && me->multi == me)
-			return CHE_FAIL;
-		mc->hash[hashval] = me->next;
-		goto delete;
-	}
-
-	while (me->next != NULL) {
-		pred = me;
-		me = me->next;
-		if (strcmp(key, me->key) == 0) {
-			if (me->multi && me->multi == me)
-				return CHE_FAIL;
-			pred->next = me->next;
-			goto delete;
-		}
-	}
-
-	return CHE_FAIL;
-
-delete:
-	status = pthread_rwlock_destroy(&me->multi_rwlock);
-	if (status)
-		fatal(status);
-	list_del(&me->multi_list);
-	ino_index_lock(mc);
-	list_del(&me->ino_index);
-	ino_index_unlock(mc);
-	free(me->key);
-	if (me->mapent)
-		free(me->mapent);
-	free(me);
-
-	return CHE_OK;
-}
-
 /* cache must be write locked by caller */
 int cache_delete(struct mapent_cache *mc, const char *key)
 {
 	struct mapent *me = NULL, *pred;
 	u_int32_t hashval = hash(key, mc->size);
-	int status, ret = CHE_OK;
-	char this[PATH_MAX];
-
-	strcpy(this, key);
+	int ret = CHE_OK;
 
 	me = mc->hash[hashval];
-	if (!me) {
-		ret = CHE_FAIL;
+	if (!me)
 		goto done;
-	}
 
 	while (me->next != NULL) {
 		pred = me;
 		me = me->next;
-		if (strcmp(this, me->key) == 0) {
+		if (strcmp(key, me->key) == 0) {
 			struct stack *s = me->stack;
-			if (me->multi && !list_empty(&me->multi_list)) {
+			if (IS_MM(me)) {
 				ret = CHE_FAIL;
 				goto done;
 			}
 			pred->next = me->next;
-			status = pthread_rwlock_destroy(&me->multi_rwlock);
-			if (status)
-				fatal(status);
 			ino_index_lock(mc);
 			list_del(&me->ino_index);
 			ino_index_unlock(mc);
@@ -1016,16 +808,13 @@ int cache_delete(struct mapent_cache *mc, const char *key)
 	if (!me)
 		goto done;
 
-	if (strcmp(this, me->key) == 0) {
+	if (strcmp(key, me->key) == 0) {
 		struct stack *s = me->stack;
-		if (me->multi && !list_empty(&me->multi_list)) {
+		if (IS_MM(me)) {
 			ret = CHE_FAIL;
 			goto done;
 		}
 		mc->hash[hashval] = me->next;
-		status = pthread_rwlock_destroy(&me->multi_rwlock);
-		if (status)
-			fatal(status);
 		ino_index_lock(mc);
 		list_del(&me->ino_index);
 		ino_index_unlock(mc);
@@ -1045,65 +834,6 @@ done:
 	return ret;
 }
 
-/* cache must be write locked by caller */
-int cache_delete_offset_list(struct mapent_cache *mc, const char *key)
-{
-	unsigned logopt = mc->ap ? mc->ap->logopt : master_get_logopt();
-	struct mapent *me;
-	struct mapent *this;
-	struct list_head *head, *next;
-	int remain = 0;
-	int status;
-
-	me = cache_lookup_distinct(mc, key);
-	if (!me)
-		return CHE_FAIL;
-
-	/* Not offset list owner */
-	if (me->multi != me)
-		return CHE_FAIL;
-
-	head = &me->multi_list;
-	next = head->next;
-	while (next != head) {
-		this = list_entry(next, struct mapent, multi_list);
-		next = next->next;
-		if (this->ioctlfd != -1) {
-			error(logopt,
-			      "active offset mount key %s", this->key);
-			return CHE_FAIL;
-		}
-	}
-
-	head = &me->multi_list;
-	next = head->next;
-	while (next != head) {
-		this = list_entry(next, struct mapent, multi_list);
-		next = next->next;
-		list_del_init(&this->multi_list);
-		this->multi = NULL;
-		debug(logopt, "deleting offset key %s", this->key);
-		status = cache_delete(mc, this->key);
-		if (status == CHE_FAIL) {
-			warn(logopt,
-			     "failed to delete offset %s", this->key);
-			this->multi = me;
-			/* TODO: add list back in */
-			remain++;
-		}
-	}
-
-	if (!remain) {
-		list_del_init(&me->multi_list);
-		me->multi = NULL;
-	}
-
-	if (remain)
-		return CHE_FAIL;
-
-	return CHE_OK;
-}
-
 void cache_release(struct map_source *map)
 {
 	struct mapent_cache *mc;
@@ -1208,113 +938,3 @@ struct mapent *cache_enumerate(struct mapent_cache *mc, struct mapent *me)
 
 	return cache_lookup_next(mc, me);
 }
-
-/*
- * Get each offset from list head under prefix.
- * Maintain traversal current position in pos for subsequent calls. 
- * Return each offset into offset.
- */
-/* cache must be read locked by caller */
-char *cache_get_offset(const char *prefix, char *offset, int start,
-			struct list_head *head, struct list_head **pos)
-{
-	struct list_head *next;
-	struct mapent *this;
-	size_t plen = strlen(prefix);
-	size_t len = 0;
-
-	if (*pos == head)
-		return NULL;
-
-	/* Find an offset */
-	*offset = '\0';
-	next = *pos ? (*pos)->next : head->next;
-	while (next != head) {
-		char *offset_start, *pstart, *pend;
-
-		this = list_entry(next, struct mapent, multi_list);
-		*pos = next;
-		next = next->next;
-
-		offset_start = &this->key[start];
-		if (strlen(offset_start) <= plen)
-			continue;
-
-		if (!strncmp(prefix, offset_start, plen)) {
-			struct mapent *np = NULL;
-			char pe[PATH_MAX + 1];
-
-			/* "/" doesn't count for root offset */
-			if (plen == 1)
-				pstart = &offset_start[plen - 1];
-			else
-				pstart = &offset_start[plen];
-
-			/* not part of this sub-tree */
-			if (*pstart != '/')
-				continue;
-
-			/* get next offset */
-			pend = pstart;
-			while (*pend++) {
-				size_t nest_pt_offset;
-
-				if (*pend != '/')
-					continue;
-
-				nest_pt_offset = start + pend - pstart;
-				if (plen > 1)
-					nest_pt_offset += plen;
-				strcpy(pe, this->key);
-				pe[nest_pt_offset] = '\0';
-
-				np = cache_lookup_distinct(this->mc, pe);
-				if (np)
-					break;
-			}
-			if (np)
-				continue;
-			len = pend - pstart - 1;
-			strncpy(offset, pstart, len);
-			offset[len] ='\0';
-			break;
-		}
-	}
-
-	/* Seek to next offset */
-	while (next != head) {
-		char *offset_start, *pstart;
-
-		this = list_entry(next, struct mapent, multi_list);
-
-		offset_start = &this->key[start];
-		if (strlen(offset_start) <= plen + len)
-			break;
-
-		/* "/" doesn't count for root offset */
-		if (plen == 1)
-			pstart = &offset_start[plen - 1];
-		else
-			pstart = &offset_start[plen];
-
-		/* not part of this sub-tree */
-		if (*pstart != '/')
-			break;
-
-		/* new offset */
-		if (!*(pstart + len + 1))
-			break;
-
-		/* compare offset */
-		if (pstart[len] != '/' ||
-		    strlen(pstart) != len ||
-		    strncmp(offset, pstart, len))
-			break;
-
-		*pos = next;
-		next = next->next;
-	}
-
-	return *offset ? offset : NULL;
-}
-
