@@ -363,7 +363,6 @@ done:
 int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *context)
 {
 	struct lookup_context *ctxt = (struct lookup_context *) context;
-	struct mapent_cache *mc;
 	char buf[MAX_ERR_BUF];
 	struct map_source *source;
 	struct mapent *me;
@@ -378,8 +377,6 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	ap->entry->current = NULL;
 	master_source_current_signal(ap->entry);
 
-	mc = source->mc;
-
 	debug(ap->logopt,
 	      MODPREFIX "looking up root=\"%s\", name=\"%s\"",
 	      ap->path, name);
@@ -400,27 +397,9 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	/* Check if we recorded a mount fail for this key anywhere */
 	me = lookup_source_mapent(ap, name, LKP_DISTINCT);
 	if (me) {
-		if (me->status >= monotonic_time(NULL)) {
-			cache_unlock(me->mc);
+		/* negative timeout has not passed, return fail */
+		if (cache_lookup_negative(me, name) == CHE_UNAVAIL)
 			return NSS_STATUS_NOTFOUND;
-		} else {
-			struct mapent_cache *smc = me->mc;
-			struct mapent *sme;
-
-			if (me->mapent)
-				cache_unlock(smc);
-			else {
-				cache_unlock(smc);
-				cache_writelock(smc);
-				sme = cache_lookup_distinct(smc, name);
-				/* Negative timeout expired for non-existent entry. */
-				if (sme && !sme->mapent) {
-					if (cache_pop_mapent(sme) == CHE_FAIL)
-						cache_delete(smc, name);
-				}
-				cache_unlock(smc);
-			}
-		}
 	}
 
 	/* If this is not here the filesystem stays busy, for some reason... */
@@ -466,14 +445,28 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	}
 
 	me = match_cached_key(ap, MODPREFIX, source, lkp_key);
-	free(lkp_key);
-	if (!me)
-		return NSS_STATUS_NOTFOUND;
 
-	if (!me->mapent)
+	if (!me) {
+		free(lkp_key);
+		return NSS_STATUS_NOTFOUND;
+	}
+
+	if (!me->mapent) {
+		free(lkp_key);
 		return NSS_STATUS_UNAVAIL;
+	}
 
 	mapent = strdup(me->mapent);
+	if (!mapent) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		error(ap->logopt, "malloc: %s", estr);
+		free(lkp_key);
+		return NSS_STATUS_UNKNOWN;
+	}
+
+	debug(ap->logopt, MODPREFIX "%s -> %s", lkp_key, mapent);
+
+	free(lkp_key);
 
 	rv = ctxt->parser->parse_mount(ap, key, key_len,
 				       mapent, ctxt->parser->context);
