@@ -380,7 +380,8 @@ static int do_init(int argc, const char *const *argv, struct parse_context *ctxt
 			if (!tmp) {
 				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 				logerr(MODPREFIX "concat_options: %s", estr);
-				free(gbl_options);
+				/* freed in concat_options */
+				ctxt->optstr = NULL;
 			} else
 				ctxt->optstr = tmp;
 		} else {
@@ -492,12 +493,16 @@ static char *concat_options(char *left, char *right)
 	char *ret;
 
 	if (left == NULL || *left == '\0') {
+		if (!right || *right == '\0')
+			return NULL;
 		ret = strdup(right);
 		free(right);
 		return ret;
 	}
 
 	if (right == NULL || *right == '\0') {
+		if (left == NULL || *left == '\0')
+			return NULL;
 		ret = strdup(left);
 		free(left);
 		return ret;
@@ -508,6 +513,8 @@ static char *concat_options(char *left, char *right)
 	if (ret == NULL) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 		logerr(MODPREFIX "malloc: %s", estr);
+		free(left);
+		free(right);
 		return NULL;
 	}
 
@@ -530,7 +537,6 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 	int nonstrict = 1;
 	int use_weight_only = ap->flags & MOUNT_FLAG_USE_WEIGHT_ONLY;
 	int rv, cur_state;
-	char *mountpoint;
 	char *what;
 	char *type;
 
@@ -624,9 +630,6 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 		}
 	}
 
-	mountpoint = alloca(namelen + 1);
-	sprintf(mountpoint, "%.*s", namelen, name);
-
 	type = ap->entry->maps->type;
 	if (type && !strcmp(type, "hosts")) {
 		if (options && *options != '\0') {
@@ -665,9 +668,16 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 		}
 	}
 
+	what = malloc(loclen + 1);
+	if (!what) {
+		char buf[MAX_ERR_BUF];
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		error(ap->logopt, MODPREFIX "malloc: %s", estr);
+		return 1;
+	}
+
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
 	if (!strcmp(fstype, "nfs") || !strcmp(fstype, "nfs4")) {
-		what = alloca(loclen + 1);
 		memcpy(what, loc, loclen);
 		what[loclen] = '\0';
 
@@ -698,15 +708,15 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 		debug(ap->logopt, MODPREFIX
 		      "mounting root %s, mountpoint %s, "
 		      "what %s, fstype %s, options %s",
-		      root, mountpoint, what, fstype, options);
+		      root, name, what, fstype, options);
 
-		rv = mount_nfs->mount_mount(ap, root, mountpoint, strlen(mountpoint),
+		rv = mount_nfs->mount_mount(ap, root, name, namelen,
 					    what, fstype, options, mount_nfs->context);
 	} else {
-		if (!loclen)
+		if (!loclen) {
+			free(what);
 			what = NULL;
-		else {
-			what = alloca(loclen + 1);
+		} else {
 			if (*loc == ':') {
 				loclen--;
 				memcpy(what, loc + 1, loclen);
@@ -720,12 +730,14 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 		debug(ap->logopt, MODPREFIX
 		      "mounting root %s, mountpoint %s, "
 		      "what %s, fstype %s, options %s",
-		      root, mountpoint, what, fstype, options);
+		      root, name, what, fstype, options);
 
 		/* Generic mount routine */
-		rv = do_mount(ap, root, mountpoint, strlen(mountpoint), what, fstype,
-			      options);
+		rv = do_mount(ap, root, name, namelen, what, fstype, options);
 	}
+	if (what)
+		free(what);
+
 	pthread_setcancelstate(cur_state, NULL);
 
 	if (nonstrict && rv)
@@ -793,46 +805,53 @@ static int check_is_multi(const char *mapent)
 }
 
 static int
-update_offset_entry(struct autofs_point *ap, const char *name,
-		    const char *m_root, int m_root_len,
-		    const char *path, const char *myoptions, const char *loc,
-		    time_t age)
+update_offset_entry(struct autofs_point *ap,
+		    struct mapent_cache *mc, struct list_head *offsets,
+		    const char *name, const char *m_root, int m_root_len,
+		    const char *m_offset, const char *myoptions,
+		    const char *loc, time_t age)
 {
-	struct map_source *source;
-	struct mapent_cache *mc;
 	char m_key[PATH_MAX + 1];
 	char m_mapent[MAPENT_MAX_LEN + 1];
-	int p_len, m_key_len, m_options_len, m_mapent_len;
+	int o_len, m_key_len, m_options_len, m_mapent_len;
+	struct mapent *me;
 	int ret;
 
-	source = ap->entry->current;
-	ap->entry->current = NULL;
-	master_source_current_signal(ap->entry);
-
-	mc = source->mc;
-
 	memset(m_mapent, 0, MAPENT_MAX_LEN + 1);
 
-	/* Internal hosts map may have loc == NULL */
-	if (!*path) {
+	if (!loc || !*loc) {
+		const char *type = ap->entry->maps->type;
+
+		/* If it's not the internal hosts map it must have a
+		 * mount location.
+		 */
+		if (!type || strcmp(type, "hosts")) {
+			error(ap->logopt,
+			      MODPREFIX "syntax error in offset %s -> %s",
+			      m_offset, loc);
+			return CHE_FAIL;
+		}
+	}
+
+	if (!*m_offset) {
 		error(ap->logopt,
-		      MODPREFIX "syntax error in offset %s -> %s", path, loc);
+		      MODPREFIX "syntax error in offset %s -> %s", m_offset, loc);
 		return CHE_FAIL;
 	}
 
-	p_len = strlen(path);
+	o_len = strlen(m_offset);
 	/* Trailing '/' causes us pain */
-	if (p_len > 1) {
-		while (p_len > 1 && path[p_len - 1] == '/')
-			p_len--;
+	if (o_len > 1) {
+		while (o_len > 1 && m_offset[o_len - 1] == '/')
+			o_len--;
 	}
-	m_key_len = m_root_len + p_len;
+	m_key_len = m_root_len + o_len;
 	if (m_key_len > PATH_MAX) {
 		error(ap->logopt, MODPREFIX "multi mount key too long");
 		return CHE_FAIL;
 	}
 	strcpy(m_key, m_root);
-	strncat(m_key, path, p_len);
+	strncat(m_key, m_offset, o_len);
 	m_key[m_key_len] = '\0';
 
 	m_options_len = 0;
@@ -858,18 +877,45 @@ update_offset_entry(struct autofs_point *ap, const char *name,
 			strcpy(m_mapent, loc);
 	}
 
+	cache_writelock(mc);
 	ret = cache_update_offset(mc, name, m_key, m_mapent, age);
+
+	me = cache_lookup_distinct(mc, m_key);
+	if (me && list_empty(&me->work)) {
+		struct list_head *last;
+
+		/* Offset entries really need to be in shortest to
+		 * longest path order. If not and the list of offsets
+		 * is large there will be a performace hit.
+		 */
+		list_for_each_prev(last, offsets) {
+			struct mapent *this;
+
+			this = list_entry(last, struct mapent, work);
+			if (me->len >= this->len) {
+				if (last->next == offsets)
+					list_add_tail(&me->work, offsets);
+				else
+					list_add_tail(&me->work, last);
+				break;
+			}
+		}
+		if (list_empty(&me->work))
+			list_add(&me->work, offsets);
+	}
+	cache_unlock(mc);
+
 	if (ret == CHE_DUPLICATE) {
 		warn(ap->logopt, MODPREFIX
-		     "syntax error or duplicate offset %s -> %s", path, loc);
+		     "syntax error or duplicate offset %s -> %s", m_offset, loc);
 		ret = CHE_OK;
 	} else if (ret == CHE_FAIL)
 		debug(ap->logopt, MODPREFIX
-		      "failed to update multi-mount offset %s -> %s", path, m_mapent);
+		      "failed to update multi-mount offset %s -> %s", m_offset, m_mapent);
 	else {
 		ret = CHE_OK;
 		debug(ap->logopt, MODPREFIX
-		      "updated multi-mount offset %s -> %s", path, m_mapent);
+		      "updated multi-mount offset %s -> %s", m_offset, m_mapent);
 	}
 
 	return ret;
@@ -960,16 +1006,13 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 			if (newopt && strstr(newopt, myoptions)) {
 				free(myoptions);
 				myoptions = newopt;
-			} else {
+			} else if (newopt) {
 				tmp = concat_options(myoptions, newopt);
 				if (!tmp) {
 					char *estr;
 					estr = strerror_r(errno, buf, MAX_ERR_BUF);
 					error(logopt, MODPREFIX
 					      "concat_options: %s", estr);
-					if (newopt)
-						free(newopt);
-					free(myoptions);
 					return 0;
 				}
 				myoptions = tmp;
@@ -1075,86 +1118,40 @@ next:
 	return (p - ent);
 }
 
-static void cleanup_multi_triggers(struct autofs_point *ap,
-			    struct mapent *me, const char *root, int start,
-			    const char *base)
-{
-	char path[PATH_MAX + 1];
-	char offset[PATH_MAX + 1];
-	char *poffset = offset;
-	struct mapent *oe;
-	struct list_head *mm_root, *pos;
-	const char o_root[] = "/";
-	const char *mm_base;
-
-	mm_root = &me->multi->multi_list;
-
-	if (!base)
-		mm_base = o_root;
-	else
-		mm_base = base;
-
-	pos = NULL;
-
-	/* Make sure "none" of the offsets have an active mount. */
-	while ((poffset = cache_get_offset(mm_base, poffset, start, mm_root, &pos))) {
-		oe = cache_lookup_offset(mm_base, poffset, start, &me->multi_list);
-		/* root offset is a special case */
-		if (!oe || !oe->mapent || (strlen(oe->key) - start) == 1)
-			continue;
-
-		strcpy(path, root);
-		strcat(path, poffset);
-		if (umount(path)) {
-			error(ap->logopt, "error recovering from mount fail");
-			error(ap->logopt, "cannot umount offset %s", path);
-		}
-	}
-
-	return;
-}
-
-static int mount_subtree(struct autofs_point *ap, struct mapent *me,
+static int mount_subtree(struct autofs_point *ap, struct mapent_cache *mc,
 			 const char *name, char *loc, char *options, void *ctxt)
 {
-	struct mapent *mm;
-	struct mapent *ro;
-	char *mm_root, *mm_base, *mm_key;
-	const char *mnt_root;
-	unsigned int mm_root_len, mnt_root_len;
-	int start, ret = 0, rv;
+	struct mapent *me;
+	int ret = 0, rv;
+
+	cache_readlock(mc);
+	me = cache_lookup_distinct(mc, name);
+	if (!me) {
+		cache_unlock(mc);
+		return 0;
+	}
 
 	rv = 0;
 
-	mm = me->multi;
-	mm_key = mm->key;
+	if (IS_MM_ROOT(me)) {
+		char key[PATH_MAX + 1];
+		struct mapent *ro;
+		size_t len;
 
-	if (*mm_key == '/') {
-		mm_root = mm_key;
-		start = strlen(mm_key);
-	} else {
-		start = strlen(ap->path) + strlen(mm_key) + 1;
-		mm_root = alloca(start + 3);
-		strcpy(mm_root, ap->path);
-		strcat(mm_root, "/");
-		strcat(mm_root, mm_key);
-	}
-	mm_root_len = strlen(mm_root);
-
-	mnt_root = mm_root;
-	mnt_root_len = mm_root_len;
-
-	if (me == me->multi) {
-		/* name = NULL */
-		/* destination = mm_root */
-		mm_base = "/";
+		len = mount_fullpath(key, PATH_MAX, ap->path, ap->len, me->key);
+		if (!len) {
+			warn(ap->logopt, "path loo long");
+			cache_unlock(mc);
+			return 1;
+		}
+		key[len] = '/';
+		key[len + 1] = 0;
 
 		/* Mount root offset if it exists */
-		ro = cache_lookup_offset(mm_base, mm_base, strlen(mm_root), &me->multi_list);
-		if (ro) {
-			char *myoptions, *ro_loc, *tmp;
+		ro = cache_lookup_distinct(me->mc, key);
+		if (ro && ro->age == MM_ROOT(me)->age) {
+			char *myoptions, *ro_loc;
 			int namelen = name ? strlen(name) : 0;
-			const char *root;
 			int ro_len;
 
 			myoptions = NULL;
@@ -1163,42 +1160,32 @@ static int mount_subtree(struct autofs_point *ap, struct mapent *me,
 			rv = parse_mapent(ro->mapent,
 				options, &myoptions, &ro_loc, ap->logopt);
 			if (!rv) {
+				cache_unlock(mc);
 				warn(ap->logopt,
 				      MODPREFIX "failed to parse root offset");
-				cache_delete_offset_list(me->mc, name);
+				cache_writelock(mc);
+				tree_mapent_delete_offsets(mc, name);
+				cache_unlock(mc);
 				return 1;
 			}
 			ro_len = 0;
 			if (ro_loc)
 				ro_len = strlen(ro_loc);
 
-			tmp = alloca(mnt_root_len + 2);
-			strcpy(tmp, mnt_root);
-			tmp[mnt_root_len] = '/';
-			tmp[mnt_root_len + 1] = '\0';
-			root = tmp;
-
-			rv = sun_mount(ap, root, name, namelen, ro_loc, ro_len, myoptions, ctxt);
+			rv = sun_mount(ap, key, name, namelen, ro_loc, ro_len, myoptions, ctxt);
 
 			free(myoptions);
 			if (ro_loc)
 				free(ro_loc);
 		}
 
-		if (ro && rv == 0) {
-			ret = mount_multi_triggers(ap, me, mnt_root, start, mm_base);
-			if (ret == -1) {
+		if (rv <= 0) {
+			ret = tree_mapent_mount_offsets(me, 1);
+			if (!ret) {
+				tree_mapent_cleanup_offsets(me);
+				cache_unlock(mc);
 				error(ap->logopt, MODPREFIX
 					 "failed to mount offset triggers");
-				cleanup_multi_triggers(ap, me, mnt_root, start, mm_base);
-				return 1;
-			}
-		} else if (rv <= 0) {
-			ret = mount_multi_triggers(ap, me, mm_root, start, mm_base);
-			if (ret == -1) {
-				error(ap->logopt, MODPREFIX
-					 "failed to mount offset triggers");
-				cleanup_multi_triggers(ap, me, mm_root, start, mm_base);
 				return 1;
 			}
 		}
@@ -1206,38 +1193,24 @@ static int mount_subtree(struct autofs_point *ap, struct mapent *me,
 		int loclen = strlen(loc);
 		int namelen = strlen(name);
 
-		mnt_root = name;
-
-		/* name = mm_root + mm_base */
-		/* destination = mm_root + mm_base = name */
-		mm_base = &me->key[start];
-
-		rv = sun_mount(ap, mnt_root, name, namelen, loc, loclen, options, ctxt);
-		if (rv == 0) {
-			ret = mount_multi_triggers(ap, me->multi, mnt_root, start, mm_base);
-			if (ret == -1) {
+		/* Mounts at nesting points must succeed for subtree
+		 * offsets to be mounted.
+		 */
+		rv = sun_mount(ap, name, name, namelen, loc, loclen, options, ctxt);
+		if (rv <= 0) {
+			ret = tree_mapent_mount_offsets(me, 1);
+			if (!ret) {
+				tree_mapent_cleanup_offsets(me);
+				cache_unlock(mc);
 				error(ap->logopt, MODPREFIX
 					 "failed to mount offset triggers");
-				cleanup_multi_triggers(ap, me, mnt_root, start, mm_base);
-				return 1;
-			}
-		} else if (rv < 0) {
-			char *mm_root_base = alloca(strlen(mm_root) + strlen(mm_base) + 1);
-	
-			strcpy(mm_root_base, mm_root);
-			strcat(mm_root_base, mm_base);
-
-			ret = mount_multi_triggers(ap, me->multi, mm_root_base, start, mm_base);
-			if (ret == -1) {
-				error(ap->logopt, MODPREFIX
-					 "failed to mount offset triggers");
-				cleanup_multi_triggers(ap, me, mm_root, start, mm_base);
 				return 1;
 			}
 		}
 	}
+	cache_unlock(mc);
 
-	/* Mount for base of tree failed */
+	/* strict mount failed */
 	if (rv > 0)
 		return rv;
 
@@ -1274,6 +1247,25 @@ static char *do_expandsunent(const char *src, const char *key,
 	return mapent;
 }
 
+static void cleanup_offset_entries(struct autofs_point *ap,
+				   struct mapent_cache *mc,
+				   struct list_head *offsets)
+{
+	struct mapent *me, *tmp;
+	int ret;
+
+	if (list_empty(offsets))
+		return;
+	cache_writelock(mc);
+	list_for_each_entry_safe(me, tmp, offsets, work) {
+		list_del(&me->work);
+		ret = cache_delete(mc, me->key);
+		if (ret != CHE_OK)
+			crit(ap->logopt, "failed to delete offset %s", me->key);
+	}
+	cache_unlock(mc);
+}
+
 /*
  * syntax is:
  *	[-options] location [location] ...
@@ -1293,7 +1285,8 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	char buf[MAX_ERR_BUF];
 	struct map_source *source;
 	struct mapent_cache *mc;
-	struct mapent *me;
+	struct mapent *me, *oe, *tmp;
+	LIST_HEAD(offsets);
 	char *pmapent, *options;
 	const char *p;
 	int mapent_len, rv = 0;
@@ -1322,7 +1315,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	if (*name == '/') {
 		cache_readlock(mc);
 		me = cache_lookup_distinct(mc, name);
-		if (me && me->multi && me->multi != me) {
+		if (me && IS_MM(me) && !IS_MM_ROOT(me)) {
 			cache_unlock(mc);
 			mapent_len = strlen(mapent) + 1;
 			pmapent = malloc(mapent_len + 1);
@@ -1381,16 +1374,12 @@ dont_expand:
 			if (mnt_options && noptions && strstr(noptions, mnt_options)) {
 				free(mnt_options);
 				mnt_options = noptions;
-			} else {
+			} else if (noptions) {
 				tmp = concat_options(mnt_options, noptions);
 				if (!tmp) {
 					char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 					error(ap->logopt,
 					      MODPREFIX "concat_options: %s", estr);
-					if (noptions)
-						free(noptions);
-					if (mnt_options)
-						free(mnt_options);
 					free(options);
 					free(pmapent);
 					return 1;
@@ -1410,15 +1399,11 @@ dont_expand:
 			if (options && mnt_options && strstr(mnt_options, options)) {
 				free(options);
 				options = mnt_options;
-			} else {
+			} else if (mnt_options) {
 				tmp = concat_options(options, mnt_options);
 				if (!tmp) {
 					char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 					error(ap->logopt, MODPREFIX "concat_options: %s", estr);
-					if (options)
-						free(options);
-					if (mnt_options)
-						free(mnt_options);
 					free(pmapent);
 					return 1;
 				}
@@ -1431,36 +1416,18 @@ dont_expand:
 	debug(ap->logopt, MODPREFIX "gathered options: %s", options);
 
 	if (check_is_multi(p)) {
-		char *m_root = NULL;
+		char m_root[PATH_MAX + 1];
 		int m_root_len;
 		time_t age;
 		int l;
 
-		/* If name starts with "/" it's a direct mount */
-		if (*name == '/') {
-			m_root_len = name_len;
-			m_root = alloca(m_root_len + 1);
-			if (!m_root) {
-				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				free(options);
-				free(pmapent);
-				logerr(MODPREFIX "alloca: %s", estr);
-				return 1;
-			}
-			strcpy(m_root, name);
-		} else {
-			m_root_len = strlen(ap->path) + name_len + 1;
-			m_root = alloca(m_root_len + 1);
-			if (!m_root) {
-				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				free(options);
-				free(pmapent);
-				logerr(MODPREFIX "alloca: %s", estr);
-				return 1;
-			}
-			strcpy(m_root, ap->path);
-			strcat(m_root, "/");
-			strcat(m_root, name);
+		m_root_len = mount_fullpath(m_root, PATH_MAX, ap->path, ap->len, name);
+		if (!m_root_len) {
+			error(ap->logopt,
+			      MODPREFIX "multi-mount root path too long");
+			free(options);
+			free(pmapent);
+			return 1;
 		}
 
 		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
@@ -1476,10 +1443,9 @@ dont_expand:
 			return 1;
 		}
 
-		cache_multi_writelock(me);
 		/* So we know we're the multi-mount root */
-		if (!me->multi)
-			me->multi = me;
+		if (!IS_MM(me))
+			MAPENT_SET_ROOT(me, tree_mapent_root(me))
 		else {
 			/*
 			 * The amd host mount type assumes the lookup name
@@ -1501,37 +1467,34 @@ dont_expand:
 			if (source->flags & MAP_FLAG_FORMAT_AMD) {
 				free(options);
 				free(pmapent);
-				cache_multi_unlock(me);
 				cache_unlock(mc);
 				pthread_setcancelstate(cur_state, NULL);
 				return 0;
 			}
 		}
-
 		age = me->age;
+		cache_unlock(mc);
 
 		/* It's a multi-mount; deal with it */
 		do {
-			char *path, *myoptions, *loc;
+			char *m_offset, *myoptions, *loc;
 			int status;
 
 			if ((*p == '"' && *(p + 1) != '/') || (*p != '"' && *p != '/')) {
 				l = 0;
-				path = dequote("/", 1, ap->logopt);
+				m_offset = dequote("/", 1, ap->logopt);
 				debug(ap->logopt,
-				      MODPREFIX "dequote(\"/\") -> %s", path);
+				      MODPREFIX "dequote(\"/\") -> %s", m_offset);
 			} else {
 				l = span_space(p, mapent_len - (p - pmapent));
-				path = sanitize_path(p, l, LKP_MULTI, ap->logopt);
+				m_offset = sanitize_path(p, l, LKP_MULTI, ap->logopt);
 				debug(ap->logopt, MODPREFIX
-				      "dequote(\"%.*s\") -> %s", l, p, path);
+				      "dequote(\"%.*s\") -> %s", l, p, m_offset);
 			}
 
-			if (!path) {
+			if (!m_offset) {
 				warn(ap->logopt, MODPREFIX "null path or out of memory");
-				cache_delete_offset_list(mc, name);
-				cache_multi_unlock(me);
-				cache_unlock(mc);
+				cleanup_offset_entries(ap, mc, &offsets);
 				free(options);
 				free(pmapent);
 				pthread_setcancelstate(cur_state, NULL);
@@ -1546,10 +1509,8 @@ dont_expand:
 
 			l = parse_mapent(p, options, &myoptions, &loc, ap->logopt);
 			if (!l) {
-				cache_delete_offset_list(mc, name);
-				cache_multi_unlock(me);
-				cache_unlock(mc);
-				free(path);
+				cleanup_offset_entries(ap, mc, &offsets);
+				free(m_offset);
 				free(options);
 				free(pmapent);
 				pthread_setcancelstate(cur_state, NULL);
@@ -1559,19 +1520,14 @@ dont_expand:
 			p += l;
 			p = skipspace(p);
 
-			master_source_current_wait(ap->entry);
-			ap->entry->current = source;
-
-			status = update_offset_entry(ap, name,
-						     m_root, m_root_len,
-						     path, myoptions, loc, age);
+			status = update_offset_entry(ap, mc, &offsets,
+						     name, m_root, m_root_len,
+						     m_offset, myoptions, loc, age);
 
 			if (status != CHE_OK) {
 				warn(ap->logopt, MODPREFIX "error adding multi-mount");
-				cache_delete_offset_list(mc, name);
-				cache_multi_unlock(me);
-				cache_unlock(mc);
-				free(path);
+				cleanup_offset_entries(ap, mc, &offsets);
+				free(m_offset);
 				free(options);
 				free(pmapent);
 				free(myoptions);
@@ -1583,24 +1539,29 @@ dont_expand:
 
 			if (loc)
 				free(loc);
-			free(path);
+			free(m_offset);
 			free(myoptions);
 		} while (*p == '/' || (*p == '"' && *(p + 1) == '/'));
 
-		/*
-		 * We've got the ordered list of multi-mount entries so go
-		 * through and remove any stale entries if this is the top
-		 * of the multi-mount and set the parent entry of each.
-		 */
-		if (me == me->multi)
-			clean_stale_multi_triggers(ap, me, NULL, NULL);
-		cache_set_parents(me);
-
-		rv = mount_subtree(ap, me, name, NULL, options, ctxt);
-
-		cache_multi_unlock(me);
+		cache_writelock(mc);
+		me = cache_lookup_distinct(mc, name);
+		if (!me) {
+			cache_unlock(mc);
+			free(options);
+			free(pmapent);
+			cleanup_offset_entries(ap, mc, &offsets);
+			pthread_setcancelstate(cur_state, NULL);
+			return 1;
+		}
+		list_for_each_entry_safe(oe, tmp, &offsets, work) {
+			if (!tree_mapent_add_node(mc, MAPENT_ROOT(me), oe))
+				error(ap->logopt, "failed to add offset %s to tree", oe->key);
+			list_del_init(&oe->work);
+		}
 		cache_unlock(mc);
 
+		rv = mount_subtree(ap, mc, name, NULL, options, ctxt);
+
 		free(options);
 		free(pmapent);
 		pthread_setcancelstate(cur_state, NULL);
@@ -1619,19 +1580,16 @@ dont_expand:
 		 */
 		cache_readlock(mc);
 		if (*name == '/' &&
-		   (me = cache_lookup_distinct(mc, name)) && me->multi) {
+		   (me = cache_lookup_distinct(mc, name)) && IS_MM(me)) {
+			cache_unlock(mc);
 			loc = strdup(p);
 			if (!loc) {
 				free(options);
 				free(pmapent);
-				cache_unlock(mc);
 				warn(ap->logopt, MODPREFIX "out of memory");
 				return 1;
 			}
-			cache_multi_writelock(me);
-			rv = mount_subtree(ap, me, name, loc, options, ctxt);
-			cache_multi_unlock(me);
-			cache_unlock(mc);
+			rv = mount_subtree(ap, mc, name, loc, options, ctxt);
 			free(loc);
 			free(options);
 			free(pmapent);
